=head1 NAME

Bloonix::Duplicity - Simple wrapper for duplicity.

=head1 SYNOPSIS

    # Backup
    buplicity->run(config => $file);

    # Restore
    buplicity->run(
        config  => $file,
        restore => $path,
        target  => $target,
        time    => $time,
    );

=head1 DESCRIPTION

B<Bloonix::Duplicity> is a light weight wrapper around duplicity that makes
it possible to configure backups easily for your server.

=head1 METHODS

=head2 C<run>

=head1 INTERNAL METHODS

The following methods are just for internal usage!

=head2 C<log>

=head2 C<run_restore>

=head2 C<run_backup>

=head2 C<get_logfile>

=head2 C<sendmail>

=head2 C<validate>

=head2 C<validate_backup>

=head2 C<validate_server>

=head2 C<validate_main>

=head2 C<validate_sendmail>

=head1 EXAMPLE CONFIG

Take a look into the configuration template in C<etc/buplicity/buplicity.conf>.

=head1 PREREQUISITES

    Config::General
    Getopt::Long
    Log::Handler
    Params::Validate
    Sys::Hostname
    Test::More

=head1 REPORT BUGS

Please report all bugs to <support(at)bloonix.de>.

=head1 AUTHOR

Jonny Schulz <support(at)bloonix.de>

=head1 COPYRIGHT

Copyright (C) 2009-2014 by Jonny Schulz. All rights reserved.

This program is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

package Bloonix::Duplicity;

use strict;
use warnings;
use Config::General;
use File::Temp;
use Log::Handler;
use Params::Validate qw//;
use Sys::Hostname;

our $VERSION = "0.4";

sub log {
    my $self = shift;
    return $self->{log};
}

sub run {
    my $class  = shift;
    my $config = $class->validate(@_);
    my $self   = bless { config => $config }, $class;

    $self->{err} = [ ];
    $self->{log} = Log::Handler->new();

    if ($config->{log}) {
        $self->log->config(config => $config->{log});
    }

    if ($config->{sendmail}) {
        $self->log->add(
            forward => {
                forward_to => sub { push @{$self->{err}}, shift->{message} },
                maxlevel   => "error",
            }
        );
    }

    $self->{target} = join("",
        "$config->{server}->{proto}://",
        $config->{server}->{username},
        $config->{server}->{password} ? ":$config->{server}->{password}" : "",
        '@',
        $config->{server}->{host},
        $config->{server}->{port} ? ":$config->{server}->{port}" : "",
        "/",
        $config->{server}->{path},
        "/",
    );

    if ($config->{restore}) {
        $self->run_restore;
    } else {
        $self->run_backup;
    }

    if (!$config->{restore}) {
        $self->sendmail;
    }
}

sub run_restore {
    my $self   = shift;
    my $config = $self->{config};
    my ($params, @paths);

    my $passphrase = "";
    my $arguments  = $config->{default_restore_opts};

    if ($config->{restore} !~ /^\//) {
        print "ERROR: please enter the full path that begins with a slash\n";
        exit 1;
    }

    foreach my $b (@{ $config->{backup} }) {
        push @paths, $b->{path};
        if ($b->{path} eq $config->{restore}) {
            $params = $b;
        }
    }

    if (!$params) {
        print "The path '$config->{restore}' is not configured in the configuration file\n";
        print "Configured file sets:\n";
        print join("\n", @paths), "\n";
        exit 1;
    }

    if (!$config->{target}) {
        print "ERROR: no target\n";
        exit 1;
    }

    if (!-e $config->{target}) {
        mkdir($config->{target}) or die $!;
    }

    if ($params->{restore_opts}) {
        # If the string begins with * the default
        # restore options must be overwritten.
        if ($params->{restore_opts} =~ s/^\s*\*\s*//) {
            $arguments = $params->{restore_opts};
        } else {
            $arguments .= " $params->{restore_opts}";
        }
    }

    if ($params->{use_gpg}) {
        if ($params->{passphrase}) {
            $passphrase = "export PASSPHRASE='$params->{passphrase}';";
        } elsif ($config->{default_passphrase}) {
            $passphrase = "export PASSPHRASE='$config->{default_passphrase}';";
        }

        if ($params->{encrypt_key}) {
            $arguments .= "--encrypt-key $params->{encrypt_key}";
        } elsif ($config->{default_encrypt_key}) {
            $arguments .= "--encrypt-key $config->{default_encrypt_key}";
        }
    }

    if ($config->{file}) {
        $arguments .= " --file-to-restore $config->{file}";
    }

    if ($config->{time}) {
        $arguments .= " -t $config->{time}";
    }

    my $p_path = $config->{restore};
    $p_path =~ s!_!__!g;
    $p_path =~ s!/!_!g;

    my $cmd = join(" ",
        $passphrase,
        $config->{duplicity},
        $arguments,
        $self->{target}.$p_path,
        $config->{target},
    );

    print "Run duplicity\n";
    print "$cmd\n";
    system($cmd);

    if ($? == 0) {
        print "Files restored to $config->{target}\n";
    }
}

sub run_backup {
    my $self   = shift;
    my $config = $self->{config};

    $SIG{__DIE__} = sub { $self->log->error(@_) };

    $self->log->notice("backup started");

    foreach my $p (@{ $config->{backup} }) {
        if ($p->{active} =~ /^(?:no|0)\z/) {
            next;
        }

        my $passphrase = "";
        my $arguments  = $config->{default_backup_opts};
        my $logfile    = `$config->{mktemp}`;
        chomp($logfile);

        if ($p->{use_gpg}) {
            if ($p->{passphrase}) {
                $passphrase = "export PASSPHRASE='$p->{passphrase}';";
            } elsif ($config->{default_passphrase}) {
                $passphrase = "export PASSPHRASE='$config->{default_passphrase}';";
            }

            if ($p->{encrypt_key}) {
                $arguments .= " --encrypt-key $p->{encrypt_key}";
            } elsif ($config->{default_encrypt_key}) {
                $arguments .= " --encrypt-key $config->{default_encrypt_key}";
            }
        }

        if ($p->{backup_opts}) {
            # If the string begins with * the default
            # backup options must be overwritten.
            if ($p->{backup_opts} =~ s/^\s*\*\s*//) {
                $arguments = $p->{backup_opts};
            } else {
                $arguments .= " $p->{backup_opts}";
            }
        }

        my $p_path = $p->{path};
        $p_path =~ s!_!__!g;
        $p_path =~ s!/!_!g;

        my $cmd = join(" ",
            $passphrase,
            $config->{duplicity},
            $arguments,
            $p->{path},
            $self->{target}.$p_path,
            "1>>$logfile",
            "2>&1",
        );

        $self->log->info("starting backup for $p->{path}");
        $self->log->debug("backup options");
        $self->log->dump(debug => $p);
        $self->log->info($cmd);

        if ($config->{test}) {
            $? = 0;
        } else {
            system($cmd);
        }

        if ($? > 0) {
            $self->log->error("backup failed for $p->{path}");
            $self->get_logfile("error", $logfile);
        } else {
            $self->get_logfile("info", $logfile);
        }

        $arguments = $config->{default_delete_opts};

        if ($p->{delete_opts}) {
            # If the string begins with * the default
            # delete options must be overwritten.
            if ($p->{delete_opts} =~ s/^\s*\*\s*//) {
                $arguments = $p->{delete_opts};
            } else {
                $arguments .= " $p->{delete_opts}";
            }
        }

        $logfile = `$config->{mktemp}`;
        chomp($logfile);

        my $del = join(" ",
            $passphrase,
            $config->{duplicity},
            $arguments,
            $self->{target}.$p_path,
            "1>>$logfile",
            "2>&1",
        );

        $self->log->info("starting deletion for $p->{path}");
        $self->log->info($del);

        if ($config->{test}) {
            $? = 0;
        } else {
            system($del);
        } 

        if ($? > 0) {
            $self->log->error("delete failed for $p->{path}");
            $self->get_logfile(error => $logfile);
        } else {
            $self->get_logfile(info => $logfile);
        }
    }

    $self->log->notice("backup finished");
}

sub get_logfile {
    my ($self, $level, $logfile) = @_;

    open my $fh, "<", $logfile or die $!;

    while (my $line = <$fh>) {
        next if $line =~ /^\s*$/;
        $self->log->log($level => $line);
    }

    close $fh;

    unlink($logfile);
}

sub sendmail {
    my $self   = shift;
    my $config = $self->{config}->{sendmail};

    if (!$config) {
        $self->log->info("sendmail not configured");
        return;
    }

    my $host    = hostname;
    my $email   = $config->{binary}." -t";
    my $subject = $config->{subject} || "backup status for $host";

    if ($config->{sender}) {
        $email .= " -f $config->{sender}";
    }

    $email .= " <<EOF\n";

    if ($config->{from}) {
        $email .= "From: $config->{from}\n";
    }

    $email .= "To: $config->{to}\n";

    if ($self->{config}->{test}) {
        $subject = "TEST $subject";
    }

    if (@{$self->{err}}) {
        $email .= "Subject: ERROR - $subject";
    } else {
        $email .= "Subject: OK - $subject";
    }

    $email .= "\n"; # data begins
    $email .= join("\n", @{$self->{err}})."\n";
    $email .= "EOF\n";

    $self->log->info("sendmail to $config->{to}");
    $self->log->debug("EMAIL\n$email");
    system($email);
}

sub validate {
    my $self = shift;

    my %options = Params::Validate::validate(@_, {
        config => {
            type => Params::Validate::SCALAR,
        },
        test => {
            type => Params::Validate::SCALAR,
            default => 0,
        },
        restore => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        target => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        file => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        time => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
    });

    if (!-r $options{config}) {
        die "Config not found '$options{config}'";
    }

    my %config = Config::General::ParseConfig($options{config});
    my $config = $self->validate_main(\%config);

    $config->{restore} = $options{restore};
    $config->{target}  = $options{target};
    $config->{test}    = $options{test};
    $config->{time}    = $options{time};
    $config->{file}    = $options{file};

    return $config;
}

sub validate_main {
    my $self = shift;

    my %options = Params::Validate::validate(@_, {
        server => {
            type => Params::Validate::HASHREF,
        },
        log => {
            type => Params::Validate::HASHREF,
            optional => 1,
        },
        sendmail => {
            type => Params::Validate::HASHREF,
            optional => 1,
        },
        backup => {
            type => Params::Validate::ARRAYREF
                    | Params::Validate::HASHREF,
        },
        duplicity => {
            type => Params::Validate::SCALAR,
            default => "/usr/bin/duplicity",
        },
        mktemp => {
            type => Params::Validate::SCALAR,
            default => "/bin/mktemp",
        },
        default_backup_opts => {
            type => Params::Validate::SCALAR,
            default => "",
        },
        default_delete_opts => {
            type => Params::Validate::SCALAR,
            default => "",
        },
        default_restore_opts => {
            type => Params::Validate::SCALAR,
            default => "",
        },
        default_encrypt_key => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        default_passphrase => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
    });

    if (!-x $options{mktemp}) {
        die "binary not found '$options{mktemp}'";
    }

    if (!-x $options{duplicity}) {
        die "binary not found '$options{duplicity}'";
    }

    $options{server} = $self->validate_server($options{server});

    if (ref($options{backup}) ne "ARRAY") {
        $options{backup} = [ $options{backup} ];
    }

    my $backup = $options{backup};

    foreach my $i (0..$#$backup) {
        $backup->[$i] = $self->validate_backup($backup->[$i]);
    }

    if ($options{sendmail}) {
        $options{sendmail} = $self->validate_sendmail($options{sendmail});
    }

    return \%options;
}

sub validate_server {
    my $self = shift;

    my %options = Params::Validate::validate(@_, {
        proto => {
            type => Params::Validate::SCALAR,
            regex => qr/^(?:ssh|scp|rsync|ftp)\z/,
            default => "scp",
        },
        host => {
            type => Params::Validate::SCALAR,
        },
        port => {
            type => Params::Validate::SCALAR,
            regex => qr/^\d+\z/,
            optional => 1,
        },
        username => {
            type => Params::Validate::SCALAR,
        },
        password => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        path => {
            type => Params::Validate::SCALAR,
            default => "",
        },
    });

    return \%options;
}

sub validate_backup {
    my $self = shift;

    my %options = Params::Validate::validate(@_, {
        path => {
            type => Params::Validate::SCALAR,
        },
        backup_opts => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        delete_opts => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        restore_opts => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        encrypt_key => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        passphrase => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        use_gpg => {
            type => Params::Validate::SCALAR,
            default => 1,
        },
        active => {
            type => Params::Validate::SCALAR,
            default => "yes",
            regex => qr/^(?:yes|no|1|0)\z/,
        },
    });

    return \%options;
}

sub validate_sendmail {
    my $self = shift;

    my %options = Params::Validate::validate(@_, {
        binary => {
            type => Params::Validate::SCALAR,
            default => "/usr/sbin/sendmail",
        },
        sender => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        from => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
        to => {
            type => Params::Validate::SCALAR,
        },
        subject => {
            type => Params::Validate::SCALAR,
            optional => 1,
        },
    });

    if (!-x $options{binary}) {
        die "sendmail not found '$options{binary}'";
    }

    return \%options;
}

1;
